// taxonomies.cc

/*--------------------------- BEGIN LICENSE BLOCK ---------------------------+
|                                                                            |
| Version: MPL 1.1/GPL 2.0/LGPL 2.1                                          |
|                                                                            |
| The contents of this file are subject to the Mozilla Public License        |
| Version 1.1 (the "License"); you may not use this file except in           |
| compliance with the License. You may obtain a copy of the License at       |
| http://www.mozilla.org/MPL/                                                |
|                                                                            |
| Software distributed under the License is distributed on an "AS IS" basis, |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License   |
| for the specific language governing rights and limitations under the       |
| License.                                                                   |
|                                                                            |
| The Original Code is the STELLA Programming Language.                      |
|                                                                            |
| The Initial Developer of the Original Code is                              |
| UNIVERSITY OF SOUTHERN CALIFORNIA, INFORMATION SCIENCES INSTITUTE          |
| 4676 Admiralty Way, Marina Del Rey, California 90292, U.S.A.               |
|                                                                            |
| Portions created by the Initial Developer are Copyright (C) 1996-2001      |
| the Initial Developer. All Rights Reserved.                                |
|                                                                            |
| Contributor(s):                                                            |
|                                                                            |
| Alternatively, the contents of this file may be used under the terms of    |
| either the GNU General Public License Version 2 or later (the "GPL"), or   |
| the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),   |
| in which case the provisions of the GPL or the LGPL are applicable instead |
| of those above. If you wish to allow use of your version of this file only |
| under the terms of either the GPL or the LGPL, and not to allow others to  |
| use your version of this file under the terms of the MPL, indicate your    |
| decision by deleting the provisions above and replace them with the notice |
| and other provisions required by the GPL or the LGPL. If you do not delete |
| the provisions above, a recipient may use your version of this file under  |
| the terms of any one of the MPL, the GPL or the LGPL.                      |
|                                                                            |
+---------------------------- END LICENSE BLOCK ----------------------------*/

#include "stella/stella-system.hh"

namespace stella {

// Spacing between postorder numbers for nodes.  Allows limited
// incremental insertions without having to renumber the whole graph.
int oNUMBERING_INTERVALo = 100;

boolean taxonomy_rootP(Taxonomy_Node* node) {
  // Return `true' if `node' is a taxonomy root node.
  return (node->parents == NIL);
}

void add_taxonomy_root(Taxonomy_Graph* graph, Taxonomy_Node* root) {
  // Add `root' as a root node to `graph'.  Only do this if
  // `root' does not have any parents and is not a `graph' root already.
  graph->roots->insert_new(root);
}

void remove_taxonomy_root(Taxonomy_Graph* graph, Taxonomy_Node* root) {
  // Remove `root' from `graph's root nodes.
  if (!(root->label == DELETED_LABEL)) {
    graph->roots->remove(root);
  }
}

void update_taxonomy_roots(Taxonomy_Graph* graph, Taxonomy_Node* node) {
  // Update `graph's roots according to `node's current state.
  if (node->parents == NIL) {
    add_taxonomy_root(graph, node);
  }
  else {
    remove_taxonomy_root(graph, node);
  }
}

Taxonomy_Node* create_taxonomy_node(Taxonomy_Graph* graph, Taxonomy_Node* node, Object* nativeobject, boolean rootP) {
  // Link the taxonomy node `node' to `nativeObject' and add it
  // to `graph'.  If it is `null', a new node will be created.  Mark it as a root
  // node if `root?' is `true'.  Return the node.
  if (!((boolean)(node))) {
    { Taxonomy_Node* self_000 = new_taxonomy_node();

      self_000->native_object = nativeobject;
      node = self_000;
    }
  }
  else {
    initialize_taxonomy_node(node);
    node->native_object = nativeobject;
  }
  add_taxonomy_node(graph, node, rootP);
  return (node);
}

void add_taxonomy_node(Taxonomy_Graph* graph, Taxonomy_Node* node, boolean rootP) {
  // Add `node' to `graph'.  Mark it as a root node if `root?'
  // is `true'.  Even though this is part of the API, it should rarely be needed,
  // since `create-taxonomy-node' does everything that's necessary.
  if (rootP) {
    add_taxonomy_root(graph, node);
  }
  graph->number_of_nodes = graph->number_of_nodes + 1;
}

void remove_taxonomy_node(Taxonomy_Graph* graph, Taxonomy_Node* node) {
  // Remove `node' from `graph' and disconnect incident links.
  { Cons* parentscopy = copy_cons_list(node->parents);
    Cons* childrencopy = copy_cons_list(node->children);

    { Taxonomy_Node* parent = NULL;
      Cons* iter_000 = parentscopy;

      while (!(iter_000 == NIL)) {
        parent = ((Taxonomy_Node*)(iter_000->value));
        iter_000 = iter_000->rest;
        unlink_taxonomy_nodes(graph, parent, node);
      }
    }
    { Taxonomy_Node* child = NULL;
      Cons* iter_001 = childrencopy;

      while (!(iter_001 == NIL)) {
        child = ((Taxonomy_Node*)(iter_001->value));
        iter_001 = iter_001->rest;
        unlink_taxonomy_nodes(graph, node, child);
      }
    }
    node->intervals = NIL;
    node->native_object = NULL;
    node->deletedP_setter(TRUE);
    remove_taxonomy_root(graph, node);
    graph->number_of_nodes = graph->number_of_nodes - 1;
  }
}

void create_taxonomy_link(Taxonomy_Graph* graph, Taxonomy_Node* parent, Taxonomy_Node* child) {
  remove_taxonomy_root(graph, child);
  child->parents = cons(parent, child->parents);
  parent->children = cons(child, parent->children);
}

void link_taxonomy_nodes(Taxonomy_Graph* graph, Taxonomy_Node* parent, Taxonomy_Node* child) {
  // Cross-link `parent' and `child' in `graph'.
  // IMPORTANT: This will automatically insert a backlink from `child' to
  // `parent', so, for maximum efficiency it should not be called a second time
  // with the arguments reversed.
  if (parent->children->membP(child)) {
    return;
  }
  if (graph->incremental_modeP) {
    graph->added_links->push(cons(parent, cons(child, NIL)));
  }
  else {
    create_taxonomy_link(graph, parent, child);
  }
}

void incrementally_link_taxonomy_nodes(Taxonomy_Graph* graph, Taxonomy_Node* parent, Taxonomy_Node* child) {
  if (parent->children->membP(child)) {
    return;
  }
  create_taxonomy_link(graph, parent, child);
  { Interval* interval = NULL;
    boolean parentexistsP = ((boolean)(parent->initial_interval));
    boolean childexistsP = ((boolean)(child->initial_interval));

    if (childexistsP) {
      if (!parentexistsP) {
        interval = create_next_postorder_interval(graph);
        parent->initial_interval = interval;
        parent->label = interval->upper_bound;
        add_taxonomy_node_interval(parent, interval);
      }
      { Interval* interval = NULL;
        Cons* iter_000 = child->intervals;

        while (!(iter_000 == NIL)) {
          interval = ((Interval*)(iter_000->value));
          iter_000 = iter_000->rest;
          propagate_foreign_interval(parent, interval);
        }
      }
    }
    else {
      if (!parentexistsP) {
        interval = create_next_postorder_interval(graph);
        parent->initial_interval = interval;
        parent->label = interval->upper_bound;
        add_taxonomy_node_interval(parent, interval);
      }
      interval = allocate_interval_for_new_leaf_node(parent);
      if (!((boolean)(interval))) {
        if (graph->renumber_if_out_of_numbersP) {
          finalize_taxonomy_graph_nonincrementally(graph);
          return;
        }
        interval = create_next_postorder_interval(graph);
      }
      child->initial_interval = interval;
      child->label = interval->upper_bound;
      add_taxonomy_node_interval(child, interval);
      parent->tree_children = parent->tree_children->concatenate(cons(child, NIL), 0);
      propagate_foreign_interval(parent, interval);
    }
  }
}

Interval* create_next_postorder_interval(Taxonomy_Graph* graph) {
  graph->number_of_foreign_interval_nodes = graph->number_of_foreign_interval_nodes + 1;
  { Interval* newinterval = new_interval();

    newinterval->lower_bound = graph->largest_postorder_number + 1;
    newinterval->upper_bound = graph->largest_postorder_number + oNUMBERING_INTERVALo;
    graph->largest_postorder_number = newinterval->upper_bound;
    return (newinterval);
  }
}

Interval* allocate_interval_for_new_leaf_node(Taxonomy_Node* parent) {
  { Interval* parentinterval = parent->initial_interval;

    if (parent->tree_children == NIL) {
      if (parentinterval->lower_bound == parentinterval->upper_bound) {
        return (NULL);
      }
      { Interval* self_000 = new_interval();

        self_000->lower_bound = parentinterval->lower_bound;
        self_000->upper_bound = floor_((parentinterval->lower_bound + parentinterval->upper_bound) / 2.0);
        { Interval* value_000 = self_000;

          return (value_000);
        }
      }
    }
    { Taxonomy_Node* rightchild = ((Taxonomy_Node*)(parent->tree_children->last()));
      Interval* rightchildinterval = rightchild->initial_interval;

      if (rightchildinterval->upper_bound >= (parentinterval->upper_bound - 1)) {
        return (NULL);
      }
      { Interval* self_001 = new_interval();

        self_001->lower_bound = rightchildinterval->upper_bound + 1;
        self_001->upper_bound = ceiling((parentinterval->upper_bound + rightchildinterval->upper_bound) / 2.0);
        { Interval* value_001 = self_001;

          return (value_001);
        }
      }
    }
  }
}

void unlink_taxonomy_nodes(Taxonomy_Graph* graph, Taxonomy_Node* parent, Taxonomy_Node* child) {
  // Remove link between `parent' and `child'.
  if (!parent->children->membP(child)) {
    return;
  }
  if (graph->incremental_modeP) {
    graph->incremental_modeP = FALSE;
  }
  parent->children = parent->children->remove(child);
  parent->tree_children = parent->tree_children->remove(child);
  child->parents = child->parents->remove(parent);
  update_taxonomy_roots(graph, child);
}

void incrementally_unlink_taxonomy_nodes(Taxonomy_Graph* graph, Taxonomy_Node* parent, Taxonomy_Node* child) {
  throw *new_stella_exception("incrementally-unlink-taxonomy-nodes: Rewrite me!");
}

Iterator* all_taxonomy_nodes(Taxonomy_Graph* graph) {
  // Given a taxonomy `graph' that has been finalized,
  // return an iterator that generates all the graph's nodes.
  { All_Purpose_Iterator* self_000 = new_all_purpose_iterator();

    self_000->iterator_object = graph->roots->copy();
    self_000->iterator_next_code = ((cpp_function_code)(&all_taxonomy_nodes_nextP));
    { All_Purpose_Iterator* value_000 = self_000;

      return (value_000);
    }
  }
}

boolean all_taxonomy_nodes_nextP(All_Purpose_Iterator* self) {
  { List* roots = ((List*)(self->iterator_object));

    if (roots->emptyP()) {
      return (FALSE);
    }
    self->value = ((Taxonomy_Node*)(roots->pop()));
    { Taxonomy_Node* child = NULL;
      Cons* iter_000 = ((Taxonomy_Node*)(self->value))->tree_children;

      while (!(iter_000 == NIL)) {
        child = ((Taxonomy_Node*)(iter_000->value));
        iter_000 = iter_000->rest;
        roots->push(child);
      }
    }
    return (TRUE);
  }
}

Taxonomy_Node* find_taxonomy_node(Taxonomy_Graph* graph, int label) {
  // Return some node in `graph' with label `label'.
  { Taxonomy_Node* node = NULL;
    Iterator* iter_000 = all_taxonomy_nodes(graph);

    while (iter_000->nextP()) {
      node = ((Taxonomy_Node*)(iter_000->value));
      if (node->label == label) {
        return (node);
      }
    }
  }
  return (NULL);
}

// Dummy label used for marking a node
int MARKER_LABEL = -99;

// Label used for marking deleted nodes
int DELETED_LABEL = -99;

boolean labeled_taxonomy_nodeP(Taxonomy_Node* node) {
  return (node->label != NULL_INTEGER);
}

boolean Taxonomy_Node::deletedP() {
  { Taxonomy_Node* self = this;

    return (self->label == DELETED_LABEL);
  }
}

boolean Taxonomy_Node::deletedP_setter(boolean value) {
  { Taxonomy_Node* self = this;

    if (value) {
      self->label = DELETED_LABEL;
    }
    else {
      self->label = NULL_INTEGER;
    }
    return (value);
  }
}

void clear_taxonomy_node(Taxonomy_Node* node) {
  // Clear all taxonomy-graph-specific information of `node',
  // but retain information about the native object and associated links.
  node->label = NULL_INTEGER;
  node->intervals = NIL;
  node->initial_interval = NULL;
  node->first_interval_lower_bound = -1;
  node->first_interval_upper_bound = -1;
  node->tree_children = NIL;
  node->total_ancestors = 1;
}

void initialize_taxonomy_node(Taxonomy_Node* node) {
  // Completely clear and initialize `node'.
  clear_taxonomy_node(node);
  node->parents = NIL;
  node->children = NIL;
  node->native_object = NULL;
}

void help_clear_taxonomy_graph(Taxonomy_Node* node) {
  if (node->label != NULL_INTEGER) {
    clear_taxonomy_node(node);
    { Taxonomy_Node* child = NULL;
      Cons* iter_000 = node->children;

      while (!(iter_000 == NIL)) {
        child = ((Taxonomy_Node*)(iter_000->value));
        iter_000 = iter_000->rest;
        help_clear_taxonomy_graph(child);
      }
    }
  }
}

void clear_taxonomy_graph(Taxonomy_Graph* graph) {
  // Clear all taxonomy-graph-specific information of `graph',
  // but retain information about the native network and associated links.
  { Taxonomy_Node* root = NULL;
    Cons* iter_000 = graph->roots->remove_deleted_members()->the_cons_list;

    while (!(iter_000 == NIL)) {
      root = ((Taxonomy_Node*)(iter_000->value));
      iter_000 = iter_000->rest;
      help_clear_taxonomy_graph(root);
    }
  }
  graph->incremental_modeP = FALSE;
  graph->number_of_foreign_interval_nodes = 0;
  graph->broken_links->clear();
}

void initialize_taxonomy_graph(Taxonomy_Graph* graph) {
  // Completely clear the taxonomy graph `graph'.
  // NOTE: Any nodes associated with `graph' will not be cleared.  If they are
  // to be reused, they have to be cleared with `initialize-taxonomy-node'.
  graph->incremental_modeP = FALSE;
  graph->renumber_if_out_of_numbersP = FALSE;
  graph->largest_postorder_number = 0;
  graph->number_of_nodes = 0;
  graph->number_of_foreign_interval_nodes = 0;
  graph->renumber_ratio = 0.1;
  graph->roots = new_list();
  graph->broken_links = new_list();
  graph->added_links = new_list();
  graph->removed_links = new_list();
}

DEFINE_STELLA_SPECIAL(oTAXONOMY_POSTORDER_NUMBERo, int , NULL_INTEGER);

int help_create_taxonomy_tree_intervals(Taxonomy_Node* self) {
  { int minchildlabel = NULL_INTEGER;
    int mylabel = NULL_INTEGER;
    int prevpostnumber = oTAXONOMY_POSTORDER_NUMBERo.get();

    if (!(self->tree_children == NIL)) {
      minchildlabel = help_create_taxonomy_tree_intervals(((Taxonomy_Node*)(self->tree_children->value)));
    }
    { Taxonomy_Node* child = NULL;
      Cons* iter_000 = self->tree_children->rest;

      while (!(iter_000 == NIL)) {
        child = ((Taxonomy_Node*)(iter_000->value));
        iter_000 = iter_000->rest;
        help_create_taxonomy_tree_intervals(child);
      }
    }
    oTAXONOMY_POSTORDER_NUMBERo.set(oTAXONOMY_POSTORDER_NUMBERo.get() + oNUMBERING_INTERVALo);
    mylabel = oTAXONOMY_POSTORDER_NUMBERo.get();
    if (self->tree_children == NIL) {
      minchildlabel = prevpostnumber + 1;
    }
    self->initial_interval = intern_taxonomy_node_interval(self, minchildlabel, mylabel);
    self->label = mylabel;
    return (minchildlabel);
  }
}

void create_taxonomy_tree_intervals(Taxonomy_Graph* graph) {
  { 
    BIND_STELLA_SPECIAL(oTAXONOMY_POSTORDER_NUMBERo, int, 0);
    { Taxonomy_Node* root = NULL;
      Cons* iter_000 = graph->roots->the_cons_list;

      while (!(iter_000 == NIL)) {
        root = ((Taxonomy_Node*)(iter_000->value));
        iter_000 = iter_000->rest;
        help_create_taxonomy_tree_intervals(root);
      }
    }
    graph->largest_postorder_number = oTAXONOMY_POSTORDER_NUMBERo.get();
  }
}

boolean Interval::memberP(int n) {
  { Interval* interval = this;

    return ((n >= interval->lower_bound) &&
        (n <= interval->upper_bound));
  }
}

boolean subinterval_ofP(Interval* subinterval, Interval* superinterval) {
  return ((subinterval->upper_bound <= superinterval->upper_bound) &&
      (subinterval->lower_bound >= superinterval->lower_bound));
}

Interval* merge_intervals(Interval* leftinterval, Interval* rightinterval) {
  { Interval* self_000 = new_interval();

    self_000->lower_bound = leftinterval->lower_bound;
    self_000->upper_bound = rightinterval->upper_bound;
    { Interval* value_000 = self_000;

      return (value_000);
    }
  }
}

Cons* all_taxonomy_node_intervals(Taxonomy_Node* node) {
  return (node->intervals);
}

Interval* intern_taxonomy_node_interval(Taxonomy_Node* node, int lowerbound, int upperbound) {
  { Interval* self_000 = new_interval();

    self_000->lower_bound = lowerbound;
    self_000->upper_bound = upperbound;
    { Interval* interval = self_000;

      add_taxonomy_node_interval(node, interval);
      return (interval);
    }
  }
}

void add_taxonomy_node_interval(Taxonomy_Node* node, Interval* interval) {
  node->intervals = cons(interval, node->intervals);
  node->first_interval_lower_bound = interval->lower_bound;
  node->first_interval_upper_bound = interval->upper_bound;
}

void remove_taxonomy_node_interval(Taxonomy_Node* node, Interval* interval) {
  { Cons* intervals = node->intervals;

    intervals = node->intervals = intervals->remove(interval);
    if (intervals == NIL) {
      node->first_interval_lower_bound = -1;
      node->first_interval_upper_bound = -1;
    }
    else {
      node->first_interval_lower_bound = ((Interval*)(intervals->value))->lower_bound;
      node->first_interval_upper_bound = ((Interval*)(intervals->value))->upper_bound;
    }
  }
}

boolean adjoin_taxonomy_node_intervalP(Taxonomy_Node* node, Interval* interval) {
  { Cons* intervalstoremove = NIL;

    { Interval* renamed_Int = NULL;
      Cons* iter_000 = node->intervals;

      while (!(iter_000 == NIL)) {
        renamed_Int = ((Interval*)(iter_000->value));
        iter_000 = iter_000->rest;
        if (renamed_Int->lower_bound < interval->lower_bound) {
          if (renamed_Int->upper_bound >= interval->upper_bound) {
            return (FALSE);
          }
          else {
            if (interval->lower_bound <= (renamed_Int->upper_bound + 1)) {
              remove_taxonomy_node_interval(node, renamed_Int);
              adjoin_taxonomy_node_intervalP(node, merge_intervals(renamed_Int, interval));
              return (TRUE);
            }
          }
        }
        else {
          if (interval->upper_bound >= renamed_Int->upper_bound) {
            intervalstoremove = cons(renamed_Int, intervalstoremove);
          }
          else {
            if (renamed_Int->lower_bound <= (interval->upper_bound + 1)) {
              remove_taxonomy_node_interval(node, renamed_Int);
              adjoin_taxonomy_node_intervalP(node, merge_intervals(interval, renamed_Int));
              return (TRUE);
            }
          }
        }
      }
    }
    { Object* renamed_Int = NULL;
      Cons* iter_001 = intervalstoremove;

      while (!(iter_001 == NIL)) {
        renamed_Int = iter_001->value;
        iter_001 = iter_001->rest;
        remove_taxonomy_node_interval(node, ((Interval*)(renamed_Int)));
      }
    }
    add_taxonomy_node_interval(node, interval);
    return (TRUE);
  }
}

void propagate_foreign_interval(Taxonomy_Node* node, Interval* interval) {
  if (node->marker == interval) {
    return;
  }
  node->marker = interval;
  if (adjoin_taxonomy_node_intervalP(node, interval)) {
    { Taxonomy_Node* parent = NULL;
      Cons* iter_000 = node->parents;

      while (!(iter_000 == NIL)) {
        parent = ((Taxonomy_Node*)(iter_000->value));
        iter_000 = iter_000->rest;
        propagate_foreign_interval(parent, interval);
      }
    }
  }
}

void uncompute_total_ancestors(Taxonomy_Node* node) {
  if (node->total_ancestors < 0) {
    return;
  }
  node->total_ancestors = -1;
  { Taxonomy_Node* child = NULL;
    Cons* iter_000 = node->children;

    while (!(iter_000 == NIL)) {
      child = ((Taxonomy_Node*)(iter_000->value));
      iter_000 = iter_000->rest;
      uncompute_total_ancestors(child);
    }
  }
}

void help_compute_total_ancestors(Taxonomy_Node* node) {
  if (node->total_ancestors >= 0) {
    return;
  }
  { int totalancestors = 0;
    int parentancestors = 0;

    { Taxonomy_Node* parent = NULL;
      Cons* iter_000 = node->parents;

      while (!(iter_000 == NIL)) {
        parent = ((Taxonomy_Node*)(iter_000->value));
        iter_000 = iter_000->rest;
        parentancestors = parent->total_ancestors;
        if (parentancestors < 0) {
          return;
        }
        totalancestors = totalancestors + parentancestors;
      }
    }
    node->total_ancestors = totalancestors + 1;
    { Taxonomy_Node* child = NULL;
      Cons* iter_001 = node->children;

      while (!(iter_001 == NIL)) {
        child = ((Taxonomy_Node*)(iter_001->value));
        iter_001 = iter_001->rest;
        help_compute_total_ancestors(child);
      }
    }
  }
}

void compute_total_ancestors(Taxonomy_Node* node) {
  uncompute_total_ancestors(node);
  help_compute_total_ancestors(node);
}

void create_taxonomy_spanning_tree(Taxonomy_Graph* graph, Taxonomy_Node* node) {
  if (node->label != NULL_INTEGER) {
    return;
  }
  node->label = MARKER_LABEL;
  { Taxonomy_Node* maxparent = NULL;
    int maxparentvalue = NULL_INTEGER;

    { Taxonomy_Node* parent = NULL;
      Cons* iter_000 = node->parents;

      while (!(iter_000 == NIL)) {
        parent = ((Taxonomy_Node*)(iter_000->value));
        iter_000 = iter_000->rest;
        if ((maxparentvalue == NULL_INTEGER) ||
            (parent->total_ancestors > maxparentvalue)) {
          maxparent = parent;
          maxparentvalue = parent->total_ancestors;
        }
      }
    }
    { Taxonomy_Node* parent = NULL;
      Cons* iter_001 = node->parents;

      while (!(iter_001 == NIL)) {
        parent = ((Taxonomy_Node*)(iter_001->value));
        iter_001 = iter_001->rest;
        if (parent == maxparent) {
          parent->tree_children = cons(node, parent->tree_children);
        }
        else {
          graph->broken_links->push(cons(parent, cons(node, NIL)));
        }
      }
    }
    { Taxonomy_Node* child = NULL;
      Cons* iter_002 = node->children;

      while (!(iter_002 == NIL)) {
        child = ((Taxonomy_Node*)(iter_002->value));
        iter_002 = iter_002->rest;
        create_taxonomy_spanning_tree(graph, child);
      }
    }
  }
}

void create_taxonomy_spanning_forest(Taxonomy_Graph* graph) {
  clear_taxonomy_graph(graph);
  graph->incremental_modeP = FALSE;
  { Taxonomy_Node* root = NULL;
    Cons* iter_000 = graph->roots->the_cons_list;

    while (!(iter_000 == NIL)) {
      root = ((Taxonomy_Node*)(iter_000->value));
      iter_000 = iter_000->rest;
      compute_total_ancestors(root);
      create_taxonomy_spanning_tree(graph, root);
    }
  }
  create_taxonomy_tree_intervals(graph);
  { Cons* link = NULL;
    Cons* iter_001 = graph->broken_links->the_cons_list;

    while (!(iter_001 == NIL)) {
      link = ((Cons*)(iter_001->value));
      iter_001 = iter_001->rest;
      propagate_foreign_interval(((Taxonomy_Node*)(link->value)), ((Taxonomy_Node*)(link->rest->value))->initial_interval);
    }
  }
  graph->broken_links->clear();
  graph->incremental_modeP = TRUE;
}

void finalize_taxonomy_graph(Taxonomy_Graph* graph) {
  // Finalize the taxonomy graph `graph'.
  if (graph->incremental_modeP &&
      ((graph->number_of_nodes > 0) &&
       ((((double)(max_(graph->number_of_foreign_interval_nodes, graph->added_links->length()))) / graph->number_of_nodes) < graph->renumber_ratio))) {
    { Cons* link = NULL;
      Cons* iter_000 = graph->added_links->the_cons_list;

      while (!(iter_000 == NIL)) {
        link = ((Cons*)(iter_000->value));
        iter_000 = iter_000->rest;
        incrementally_link_taxonomy_nodes(graph, ((Taxonomy_Node*)(link->value)), ((Taxonomy_Node*)(link->rest->value)));
      }
    }
    graph->added_links->clear();
  }
  else {
    graph->incremental_modeP = FALSE;
    { Cons* link = NULL;
      Cons* iter_001 = graph->added_links->the_cons_list;

      while (!(iter_001 == NIL)) {
        link = ((Cons*)(iter_001->value));
        iter_001 = iter_001->rest;
        link_taxonomy_nodes(graph, ((Taxonomy_Node*)(link->value)), ((Taxonomy_Node*)(link->rest->value)));
      }
    }
    graph->added_links->clear();
    create_taxonomy_spanning_forest(graph);
  }
}

void finalize_taxonomy_graph_nonincrementally(Taxonomy_Graph* graph) {
  graph->incremental_modeP = FALSE;
  finalize_taxonomy_graph(graph);
}

boolean taxonomy_subnode_ofP(Taxonomy_Node* sub_node, Taxonomy_Node* super_node) {
  // Return TRUE if `sub-node' is a descendant of `super-node'.
  { int sub_node_label = sub_node->label;

    { Interval* interval = NULL;
      Cons* iter_000 = super_node->intervals;

      while (!(iter_000 == NIL)) {
        interval = ((Interval*)(iter_000->value));
        iter_000 = iter_000->rest;
        if ((sub_node_label >= interval->lower_bound) &&
            (sub_node_label <= interval->upper_bound)) {
          return (TRUE);
        }
      }
    }
    return (FALSE);
  }
}

boolean slow_taxonomy_subnode_ofP(Taxonomy_Node* subnode, Taxonomy_Node* supernode) {
  if (subnode == supernode) {
    return (TRUE);
  }
  { Taxonomy_Node* parent = NULL;
    Cons* iter_000 = subnode->parents;

    while (!(iter_000 == NIL)) {
      parent = ((Taxonomy_Node*)(iter_000->value));
      iter_000 = iter_000->rest;
      if (slow_taxonomy_subnode_ofP(parent, supernode)) {
        return (TRUE);
      }
    }
  }
  return (FALSE);
}

void print_taxonomy_tree(Taxonomy_Node* node, int indent, Output_Stream* stream) {
  { int i = NULL_INTEGER;
    int iter_000 = 1;
    int upper_bound_000 = indent;
    boolean unboundedP_000 = upper_bound_000 == NULL_INTEGER;

    while (unboundedP_000 ||
        (iter_000 <= upper_bound_000)) {
      i = iter_000;
      iter_000 = iter_000 + 1;
      *(stream->native_stream) << " ";
    }
  }
  *(stream->native_stream) << node << endl;
  { Taxonomy_Node* child = NULL;
    Cons* iter_001 = node->children;

    while (!(iter_001 == NIL)) {
      child = ((Taxonomy_Node*)(iter_001->value));
      iter_001 = iter_001->rest;
      print_taxonomy_tree(child, indent + 2, stream);
    }
  }
}

void print_taxonomy_graph(Taxonomy_Graph* graph, Output_Stream* stream) {
  { Taxonomy_Node* root = NULL;
    Cons* iter_000 = graph->roots->the_cons_list;

    while (!(iter_000 == NIL)) {
      root = ((Taxonomy_Node*)(iter_000->value));
      iter_000 = iter_000->rest;
      print_taxonomy_tree(root, 0, stream);
      *(stream->native_stream) << endl;
    }
  }
}

void print_taxonomy_spanning_tree(Taxonomy_Node* node, int indent, Output_Stream* stream) {
  { int i = NULL_INTEGER;
    int iter_000 = 1;
    int upper_bound_000 = indent;
    boolean unboundedP_000 = upper_bound_000 == NULL_INTEGER;

    while (unboundedP_000 ||
        (iter_000 <= upper_bound_000)) {
      i = iter_000;
      iter_000 = iter_000 + 1;
      *(stream->native_stream) << " ";
    }
  }
  *(stream->native_stream) << node << endl;
  { Taxonomy_Node* child = NULL;
    Cons* iter_001 = node->tree_children;

    while (!(iter_001 == NIL)) {
      child = ((Taxonomy_Node*)(iter_001->value));
      iter_001 = iter_001->rest;
      print_taxonomy_spanning_tree(child, indent + 2, stream);
    }
  }
}

void print_taxonomy_spanning_forest(Taxonomy_Graph* graph, Output_Stream* stream) {
  { Taxonomy_Node* root = NULL;
    Cons* iter_000 = graph->roots->the_cons_list;

    while (!(iter_000 == NIL)) {
      root = ((Taxonomy_Node*)(iter_000->value));
      iter_000 = iter_000->rest;
      print_taxonomy_spanning_tree(root, 0, stream);
      *(stream->native_stream) << endl;
    }
  }
}

void startup_taxonomies() {
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, oSTELLA_MODULEo);
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    if (current_startup_time_phaseP(2)) {
      SYM_TAXONOMIES_STELLA_STARTUP_TAXONOMIES = ((Symbol*)(intern_rigid_symbol_wrt_module("STARTUP-TAXONOMIES", NULL, 0)));
      SYM_TAXONOMIES_STELLA_METHOD_STARTUP_CLASSNAME = ((Symbol*)(intern_rigid_symbol_wrt_module("METHOD-STARTUP-CLASSNAME", NULL, 0)));
    }
    if (current_startup_time_phaseP(6)) {
      finalize_classes();
    }
    if (current_startup_time_phaseP(7)) {
      define_function_object("TAXONOMY-ROOT?", "(DEFUN (TAXONOMY-ROOT? BOOLEAN) ((NODE TAXONOMY-NODE)) :DOCUMENTATION \"Return `true' if `node' is a taxonomy root node.\" :GLOBALLY-INLINE? TRUE (RETURN (EMPTY? (PARENTS NODE))))", ((cpp_function_code)(&taxonomy_rootP)), NULL);
      define_function_object("ADD-TAXONOMY-ROOT", "(DEFUN ADD-TAXONOMY-ROOT ((GRAPH TAXONOMY-GRAPH) (ROOT TAXONOMY-NODE)) :DOCUMENTATION \"Add `root' as a root node to `graph'.  Only do this if\n`root' does not have any parents and is not a `graph' root already.\" :PUBLIC? TRUE)", ((cpp_function_code)(&add_taxonomy_root)), NULL);
      define_function_object("REMOVE-TAXONOMY-ROOT", "(DEFUN REMOVE-TAXONOMY-ROOT ((GRAPH TAXONOMY-GRAPH) (ROOT TAXONOMY-NODE)) :DOCUMENTATION \"Remove `root' from `graph's root nodes.\" :PUBLIC? TRUE)", ((cpp_function_code)(&remove_taxonomy_root)), NULL);
      define_function_object("UPDATE-TAXONOMY-ROOTS", "(DEFUN UPDATE-TAXONOMY-ROOTS ((GRAPH TAXONOMY-GRAPH) (NODE TAXONOMY-NODE)) :DOCUMENTATION \"Update `graph's roots according to `node's current state.\" :PUBLIC? TRUE)", ((cpp_function_code)(&update_taxonomy_roots)), NULL);
      define_function_object("CREATE-TAXONOMY-NODE", "(DEFUN (CREATE-TAXONOMY-NODE TAXONOMY-NODE) ((GRAPH TAXONOMY-GRAPH) (NODE TAXONOMY-NODE) (NATIVEOBJECT OBJECT) (ROOT? BOOLEAN)) :DOCUMENTATION \"Link the taxonomy node `node' to `nativeObject' and add it\nto `graph'.  If it is `null', a new node will be created.  Mark it as a root\nnode if `root?' is `true'.  Return the node.\" :PUBLIC? TRUE)", ((cpp_function_code)(&create_taxonomy_node)), NULL);
      define_function_object("ADD-TAXONOMY-NODE", "(DEFUN ADD-TAXONOMY-NODE ((GRAPH TAXONOMY-GRAPH) (NODE TAXONOMY-NODE) (ROOT? BOOLEAN)) :DOCUMENTATION \"Add `node' to `graph'.  Mark it as a root node if `root?'\nis `true'.  Even though this is part of the API, it should rarely be needed,\nsince `create-taxonomy-node' does everything that's necessary.\" :PUBLIC? TRUE)", ((cpp_function_code)(&add_taxonomy_node)), NULL);
      define_function_object("REMOVE-TAXONOMY-NODE", "(DEFUN REMOVE-TAXONOMY-NODE ((GRAPH TAXONOMY-GRAPH) (NODE TAXONOMY-NODE)) :PUBLIC? TRUE :DOCUMENTATION \"Remove `node' from `graph' and disconnect incident links.\")", ((cpp_function_code)(&remove_taxonomy_node)), NULL);
      define_function_object("CREATE-TAXONOMY-LINK", "(DEFUN CREATE-TAXONOMY-LINK ((GRAPH TAXONOMY-GRAPH) (PARENT TAXONOMY-NODE) (CHILD TAXONOMY-NODE)) :PUBLIC? TRUE)", ((cpp_function_code)(&create_taxonomy_link)), NULL);
      define_function_object("LINK-TAXONOMY-NODES", "(DEFUN LINK-TAXONOMY-NODES ((GRAPH TAXONOMY-GRAPH) (PARENT TAXONOMY-NODE) (CHILD TAXONOMY-NODE)) :DOCUMENTATION \"Cross-link `parent' and `child' in `graph'.\nIMPORTANT: This will automatically insert a backlink from `child' to\n`parent', so, for maximum efficiency it should not be called a second time\nwith the arguments reversed.\" :PUBLIC? TRUE)", ((cpp_function_code)(&link_taxonomy_nodes)), NULL);
      define_function_object("INCREMENTALLY-LINK-TAXONOMY-NODES", "(DEFUN INCREMENTALLY-LINK-TAXONOMY-NODES ((GRAPH TAXONOMY-GRAPH) (PARENT TAXONOMY-NODE) (CHILD TAXONOMY-NODE)))", ((cpp_function_code)(&incrementally_link_taxonomy_nodes)), NULL);
      define_function_object("CREATE-NEXT-POSTORDER-INTERVAL", "(DEFUN (CREATE-NEXT-POSTORDER-INTERVAL INTERVAL) ((GRAPH TAXONOMY-GRAPH)))", ((cpp_function_code)(&create_next_postorder_interval)), NULL);
      define_function_object("ALLOCATE-INTERVAL-FOR-NEW-LEAF-NODE", "(DEFUN (ALLOCATE-INTERVAL-FOR-NEW-LEAF-NODE INTERVAL) ((PARENT TAXONOMY-NODE)))", ((cpp_function_code)(&allocate_interval_for_new_leaf_node)), NULL);
      define_function_object("UNLINK-TAXONOMY-NODES", "(DEFUN UNLINK-TAXONOMY-NODES ((GRAPH TAXONOMY-GRAPH) (PARENT TAXONOMY-NODE) (CHILD TAXONOMY-NODE)) :DOCUMENTATION \"Remove link between `parent' and `child'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&unlink_taxonomy_nodes)), NULL);
      define_function_object("INCREMENTALLY-UNLINK-TAXONOMY-NODES", "(DEFUN INCREMENTALLY-UNLINK-TAXONOMY-NODES ((GRAPH TAXONOMY-GRAPH) (PARENT TAXONOMY-NODE) (CHILD TAXONOMY-NODE)))", ((cpp_function_code)(&incrementally_unlink_taxonomy_nodes)), NULL);
      define_function_object("ALL-TAXONOMY-NODES", "(DEFUN (ALL-TAXONOMY-NODES (ITERATOR OF TAXONOMY-NODE)) ((GRAPH TAXONOMY-GRAPH)) :DOCUMENTATION \"Given a taxonomy `graph' that has been finalized,\nreturn an iterator that generates all the graph's nodes.\" :PUBLIC? TRUE)", ((cpp_function_code)(&all_taxonomy_nodes)), NULL);
      define_function_object("ALL-TAXONOMY-NODES-NEXT?", "(DEFUN (ALL-TAXONOMY-NODES-NEXT? BOOLEAN) ((SELF (ALL-PURPOSE-ITERATOR OF TAXONOMY-NODE))))", ((cpp_function_code)(&all_taxonomy_nodes_nextP)), NULL);
      define_function_object("FIND-TAXONOMY-NODE", "(DEFUN (FIND-TAXONOMY-NODE TAXONOMY-NODE) ((GRAPH TAXONOMY-GRAPH) (LABEL INTEGER)) :DOCUMENTATION \"Return some node in `graph' with label `label'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&find_taxonomy_node)), NULL);
      define_function_object("LABELED-TAXONOMY-NODE?", "(DEFUN (LABELED-TAXONOMY-NODE? BOOLEAN) ((NODE TAXONOMY-NODE)) :GLOBALLY-INLINE? TRUE (RETURN (DEFINED? (LABEL NODE))))", ((cpp_function_code)(&labeled_taxonomy_nodeP)), NULL);
      define_method_object("(DEFMETHOD (DELETED? BOOLEAN) ((SELF TAXONOMY-NODE)) :GLOBALLY-INLINE? TRUE (RETURN (EQL? (LABEL SELF) DELETED-LABEL)))", ((cpp_method_code)(&Taxonomy_Node::deletedP)), ((cpp_method_code)(NULL)));
      define_method_object("(DEFMETHOD (DELETED?-SETTER BOOLEAN) ((SELF TAXONOMY-NODE) (VALUE BOOLEAN)))", ((cpp_method_code)(&Taxonomy_Node::deletedP_setter)), ((cpp_method_code)(NULL)));
      define_function_object("CLEAR-TAXONOMY-NODE", "(DEFUN CLEAR-TAXONOMY-NODE ((NODE TAXONOMY-NODE)) :DOCUMENTATION \"Clear all taxonomy-graph-specific information of `node',\nbut retain information about the native object and associated links.\" :PUBLIC? TRUE)", ((cpp_function_code)(&clear_taxonomy_node)), NULL);
      define_function_object("INITIALIZE-TAXONOMY-NODE", "(DEFUN INITIALIZE-TAXONOMY-NODE ((NODE TAXONOMY-NODE)) :DOCUMENTATION \"Completely clear and initialize `node'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&initialize_taxonomy_node)), NULL);
      define_function_object("HELP-CLEAR-TAXONOMY-GRAPH", "(DEFUN HELP-CLEAR-TAXONOMY-GRAPH ((NODE TAXONOMY-NODE)))", ((cpp_function_code)(&help_clear_taxonomy_graph)), NULL);
      define_function_object("CLEAR-TAXONOMY-GRAPH", "(DEFUN CLEAR-TAXONOMY-GRAPH ((GRAPH TAXONOMY-GRAPH)) :DOCUMENTATION \"Clear all taxonomy-graph-specific information of `graph',\nbut retain information about the native network and associated links.\" :PUBLIC? TRUE)", ((cpp_function_code)(&clear_taxonomy_graph)), NULL);
      define_function_object("INITIALIZE-TAXONOMY-GRAPH", "(DEFUN INITIALIZE-TAXONOMY-GRAPH ((GRAPH TAXONOMY-GRAPH)) :DOCUMENTATION \"Completely clear the taxonomy graph `graph'.\nNOTE: Any nodes associated with `graph' will not be cleared.  If they are\nto be reused, they have to be cleared with `initialize-taxonomy-node'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&initialize_taxonomy_graph)), NULL);
      define_function_object("HELP-CREATE-TAXONOMY-TREE-INTERVALS", "(DEFUN (HELP-CREATE-TAXONOMY-TREE-INTERVALS INTEGER) ((SELF TAXONOMY-NODE)))", ((cpp_function_code)(&help_create_taxonomy_tree_intervals)), NULL);
      define_function_object("CREATE-TAXONOMY-TREE-INTERVALS", "(DEFUN CREATE-TAXONOMY-TREE-INTERVALS ((GRAPH TAXONOMY-GRAPH)))", ((cpp_function_code)(&create_taxonomy_tree_intervals)), NULL);
      define_method_object("(DEFMETHOD (MEMBER? BOOLEAN) ((INTERVAL INTERVAL) (N INTEGER)) :GLOBALLY-INLINE? TRUE (RETURN (AND (>= N (LOWER-BOUND INTERVAL)) (<= N (UPPER-BOUND INTERVAL)))))", ((cpp_method_code)(&Interval::memberP)), ((cpp_method_code)(NULL)));
      define_function_object("SUBINTERVAL-OF?", "(DEFUN (SUBINTERVAL-OF? BOOLEAN) ((SUBINTERVAL INTERVAL) (SUPERINTERVAL INTERVAL)) :GLOBALLY-INLINE? TRUE (RETURN (AND (<= (UPPER-BOUND SUBINTERVAL) (UPPER-BOUND SUPERINTERVAL)) (>= (LOWER-BOUND SUBINTERVAL) (LOWER-BOUND SUPERINTERVAL)))))", ((cpp_function_code)(&subinterval_ofP)), NULL);
      define_function_object("MERGE-INTERVALS", "(DEFUN (MERGE-INTERVALS INTERVAL) ((LEFTINTERVAL INTERVAL) (RIGHTINTERVAL INTERVAL)) :GLOBALLY-INLINE? TRUE (RETURN (NEW INTERVAL :LOWER-BOUND (LOWER-BOUND LEFTINTERVAL) :UPPER-BOUND (UPPER-BOUND RIGHTINTERVAL))))", ((cpp_function_code)(&merge_intervals)), NULL);
      define_function_object("ALL-TAXONOMY-NODE-INTERVALS", "(DEFUN (ALL-TAXONOMY-NODE-INTERVALS (CONS OF INTERVAL)) ((NODE TAXONOMY-NODE)) :GLOBALLY-INLINE? TRUE (RETURN (INTERVALS NODE)))", ((cpp_function_code)(&all_taxonomy_node_intervals)), NULL);
      define_function_object("INTERN-TAXONOMY-NODE-INTERVAL", "(DEFUN (INTERN-TAXONOMY-NODE-INTERVAL INTERVAL) ((NODE TAXONOMY-NODE) (LOWERBOUND INTEGER) (UPPERBOUND INTEGER)))", ((cpp_function_code)(&intern_taxonomy_node_interval)), NULL);
      define_function_object("ADD-TAXONOMY-NODE-INTERVAL", "(DEFUN ADD-TAXONOMY-NODE-INTERVAL ((NODE TAXONOMY-NODE) (INTERVAL INTERVAL)))", ((cpp_function_code)(&add_taxonomy_node_interval)), NULL);
      define_function_object("REMOVE-TAXONOMY-NODE-INTERVAL", "(DEFUN REMOVE-TAXONOMY-NODE-INTERVAL ((NODE TAXONOMY-NODE) (INTERVAL INTERVAL)))", ((cpp_function_code)(&remove_taxonomy_node_interval)), NULL);
      define_function_object("ADJOIN-TAXONOMY-NODE-INTERVAL?", "(DEFUN (ADJOIN-TAXONOMY-NODE-INTERVAL? BOOLEAN) ((NODE TAXONOMY-NODE) (INTERVAL INTERVAL)))", ((cpp_function_code)(&adjoin_taxonomy_node_intervalP)), NULL);
      define_function_object("PROPAGATE-FOREIGN-INTERVAL", "(DEFUN PROPAGATE-FOREIGN-INTERVAL ((NODE TAXONOMY-NODE) (INTERVAL INTERVAL)))", ((cpp_function_code)(&propagate_foreign_interval)), NULL);
      define_function_object("UNCOMPUTE-TOTAL-ANCESTORS", "(DEFUN UNCOMPUTE-TOTAL-ANCESTORS ((NODE TAXONOMY-NODE)))", ((cpp_function_code)(&uncompute_total_ancestors)), NULL);
      define_function_object("HELP-COMPUTE-TOTAL-ANCESTORS", "(DEFUN HELP-COMPUTE-TOTAL-ANCESTORS ((NODE TAXONOMY-NODE)))", ((cpp_function_code)(&help_compute_total_ancestors)), NULL);
      define_function_object("COMPUTE-TOTAL-ANCESTORS", "(DEFUN COMPUTE-TOTAL-ANCESTORS ((NODE TAXONOMY-NODE)))", ((cpp_function_code)(&compute_total_ancestors)), NULL);
      define_function_object("CREATE-TAXONOMY-SPANNING-TREE", "(DEFUN CREATE-TAXONOMY-SPANNING-TREE ((GRAPH TAXONOMY-GRAPH) (NODE TAXONOMY-NODE)))", ((cpp_function_code)(&create_taxonomy_spanning_tree)), NULL);
      define_function_object("CREATE-TAXONOMY-SPANNING-FOREST", "(DEFUN CREATE-TAXONOMY-SPANNING-FOREST ((GRAPH TAXONOMY-GRAPH)))", ((cpp_function_code)(&create_taxonomy_spanning_forest)), NULL);
      define_function_object("FINALIZE-TAXONOMY-GRAPH", "(DEFUN FINALIZE-TAXONOMY-GRAPH ((GRAPH TAXONOMY-GRAPH)) :DOCUMENTATION \"Finalize the taxonomy graph `graph'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&finalize_taxonomy_graph)), NULL);
      define_function_object("FINALIZE-TAXONOMY-GRAPH-NONINCREMENTALLY", "(DEFUN FINALIZE-TAXONOMY-GRAPH-NONINCREMENTALLY ((GRAPH TAXONOMY-GRAPH)))", ((cpp_function_code)(&finalize_taxonomy_graph_nonincrementally)), NULL);
      define_function_object("TAXONOMY-SUBNODE-OF?", "(DEFUN (TAXONOMY-SUBNODE-OF? BOOLEAN) ((SUB-NODE TAXONOMY-NODE) (SUPER-NODE TAXONOMY-NODE)) :DOCUMENTATION \"Return TRUE if `sub-node' is a descendant of `super-node'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&taxonomy_subnode_ofP)), NULL);
      define_function_object("SLOW-TAXONOMY-SUBNODE-OF?", "(DEFUN (SLOW-TAXONOMY-SUBNODE-OF? BOOLEAN) ((SUBNODE TAXONOMY-NODE) (SUPERNODE TAXONOMY-NODE)))", ((cpp_function_code)(&slow_taxonomy_subnode_ofP)), NULL);
      define_function_object("PRINT-TAXONOMY-TREE", "(DEFUN PRINT-TAXONOMY-TREE ((NODE TAXONOMY-NODE) (INDENT INTEGER) (STREAM OUTPUT-STREAM)))", ((cpp_function_code)(&print_taxonomy_tree)), NULL);
      define_function_object("PRINT-TAXONOMY-GRAPH", "(DEFUN PRINT-TAXONOMY-GRAPH ((GRAPH TAXONOMY-GRAPH) (STREAM OUTPUT-STREAM)))", ((cpp_function_code)(&print_taxonomy_graph)), NULL);
      define_function_object("PRINT-TAXONOMY-SPANNING-TREE", "(DEFUN PRINT-TAXONOMY-SPANNING-TREE ((NODE TAXONOMY-NODE) (INDENT INTEGER) (STREAM OUTPUT-STREAM)))", ((cpp_function_code)(&print_taxonomy_spanning_tree)), NULL);
      define_function_object("PRINT-TAXONOMY-SPANNING-FOREST", "(DEFUN PRINT-TAXONOMY-SPANNING-FOREST ((GRAPH TAXONOMY-GRAPH) (STREAM OUTPUT-STREAM)))", ((cpp_function_code)(&print_taxonomy_spanning_forest)), NULL);
      define_function_object("STARTUP-TAXONOMIES", "(DEFUN STARTUP-TAXONOMIES () :PUBLIC? TRUE)", ((cpp_function_code)(&startup_taxonomies)), NULL);
      { Method_Slot* function = lookup_function(SYM_TAXONOMIES_STELLA_STARTUP_TAXONOMIES);

        set_dynamic_slot_value(function->dynamic_slots, SYM_TAXONOMIES_STELLA_METHOD_STARTUP_CLASSNAME, wrap_string("Startup-Taxonomies"), NULL_STRING_WRAPPER);
      }
    }
    if (current_startup_time_phaseP(8)) {
      finalize_slots();
      cleanup_unfinalized_classes();
    }
    if (current_startup_time_phaseP(9)) {
      define_stella_global_variable_from_stringified_source("(DEFGLOBAL *NUMBERING-INTERVAL* INTEGER 100 :PUBLIC? TRUE :DOCUMENTATION \"Spacing between postorder numbers for nodes.  Allows limited\nincremental insertions without having to renumber the whole graph.\")");
      define_stella_global_variable_from_stringified_source("(DEFCONSTANT MARKER-LABEL INTEGER -99 :DOCUMENTATION \"Dummy label used for marking a node\")");
      define_stella_global_variable_from_stringified_source("(DEFCONSTANT DELETED-LABEL INTEGER -99 :DOCUMENTATION \"Label used for marking deleted nodes\")");
      define_stella_global_variable_from_stringified_source("(DEFSPECIAL *TAXONOMY-POSTORDER-NUMBER* INTEGER NULL)");
    }
  }
}

Symbol* SYM_TAXONOMIES_STELLA_STARTUP_TAXONOMIES = NULL;

Symbol* SYM_TAXONOMIES_STELLA_METHOD_STARTUP_CLASSNAME = NULL;

} // end of namespace stella
